#!/bin/bash
# Purpose: _Fully_ rip a DVD or BD to MKV (DVD only for now).
#
# LICENCE: GNU GPL v3 or later
#
# Dependencies (needed in PATH):
# — isosize (part of util-linux)                    : get sector size and count of the disc (unless -n)
# — dd, ddrescue                                    : copy image of disc                    (unless -n)
# — HandBrakeCLI (part of HandBrake)                : get disc metadata and rip chapters of each title to MKV
# — mplayer                                         : detect black borders of each title
# — ffmpeg                                          : extract a single chosen frame from an MKV
# — identify, montage, convert (part of ImageMagick): get extracted frame metadata, create visual timelines
# — czkawka_cli (part of Czkawka)                   : compare timelines and detect duplicate chapters
# — mkvmerge                                        : get MKV metadata and assemble chapters into a final MKV
# — sed, awk, grep, sort, uniq, head, cut, tr, wc   : data manipulation
# — dirname, tee, ln, mktemp, cat, touch, mv, rm    : file manipulation
# — jq                                              : for parsing JSON and extracting metadata
# — bc                                              : arbitrary precision calculator
# — date                                            : conversion between timestamps and hh:mm:ss.nnnnnnnnn
#
# Options:
# [-h                  : show this help                                                                         ]
# [-?                  : display metadata and exit (implies -n)                                                 ]
# [-l lang             : default language for audio (and maybe video?); default: fra                            ]
# [-L lang1,lang2,…    : all languages to retain for audio and subtitles; default: eng,fra                      ]
# [-v debug|info|quiet : verbosity; default: info                                                               ]
# [-t temp_path        : location with at least 2×disc size, to store temporary files; default: /tmp/bdvdrip-xxx]
# [-i input_path       : path of the disc device or ISO; default: /dev/sr0                                      ]
# [-o output_path      : path of target video file to write (`.mkv` extension); default: ./output.mkv           ]
# [-n                  : do Not dump the input_path to an ISO file; symlink instead                             ]
# [-g MAX_XxMAX_Y      : maximum geometry of the encoded video (after border removal); default: 1280x720 (720p) ]
#
# IMPORTANT: https://unix.stackexchange.com/a/493393
#
# Technical references:
# — http://mod16.org/hurfdurf/?p=8
# — https://www.reddit.com/r/mkvtoolnix/comments/11nbfy0/tutorial_mediumlinked_tiny_segmented_mkvs_with/
# — https://www.reddit.com/r/handbrake/comments/bhqxve/handbrake_settings_explained/

trap 'exit 100' SIGINT

while [ $# -gt 0 ]; do
  case "$1" in
  -h) sed -nr '2,/^$/s/^# ?//p' "$0"; exit 0 ;;
  -\?) NO_DD=true; DEBUG=quiet; SHOW=true ;;
  -l) DEFAULT_LANG="$2"; shift ;;
  -L) ALL_LANGS="$2"; shift ;;
  -v) DEBUG="$2"; shift ;;
  -t) BUILD="$2"; shift ;;
  -i) DVD="$2"; shift ;;
  -o) TARGET="$2"; shift ;;
  -n) NO_DD=true ;;
  -g) IFS=x read -r MAXX MAXY <<<"$2"; shift ;;
  *) { echo "Unknown option: $1"; sed -n '2,/^$/s/^..//p' "$0"; } >&2; exit 1 ;;
  esac
  shift
done

DEFAULT_LANG=${DEFAULT_LANG:-fra}
ALL_LANGS=${ALL_LANGS:-eng,fra}
KEEP_ALL=${BUILD:+true}
BUILD="${BUILD:-$(mktemp -d /tmp/bdvdrip-XXXXXX)}"
[ -d "$BUILD" ] && [ -w "$BUILD" ] || { echo "$BUILD is not a writeable directory"; exit 1; }
DVD="${DVD:-/dev/sr0}"
[ -r "$DVD" ] && ! [ -d "$DVD" ] || [ -f "$BUILD/.iso" ] || { echo "$DVD is not a useable readable input"; exit 1; }
TARGET="${TARGET:-./output.mkv}"
[ -d "$(dirname "$TARGET")" ] && [ -w "$(dirname "$TARGET")" ] || { echo "$(dirname "$TARGET") is not a writeable directory"; exit 1; }
if ! [[ "$MAXX" =~ ^[0-9]+$ ]]; then MAXX=1280; fi
if ! [[ "$MAXY" =~ ^[0-9]+$ ]]; then MAXY=720; fi

exec 3>&2
function log_and_run() {
  local getOutput
  if [ "$1" == '-o' ]; then
    getOutput=1; shift
  fi
  local step="$1"; shift
  if [ "$DEBUG" != quiet ]; then
    printf '\n>> %s\n>>' "$step" >&3
    printf ' %q' "$@" >&3
    printf '\n' >&3
  fi
  case "$getOutput.$DEBUG" in
  1.debug) "$@" 1> >(tee >(cat >&3)) 2> >(tee >(cat >&3) >&2) ;;
  1.*)     "$@" ;;
  .debug)  "$@" 1>&3 2>&3 ;;
  .*)      "$@" &>/dev/null ;;
  esac
}

# copy DVD

if [ ! -f "$BUILD/.iso" ]; then
  if [ -n "$NO_DD" ]; then
    RM_ISO=${SHOW:+true}
    ln -s "$DVD" "$BUILD/.iso"
  else
    read -r sectcount sectsize < <(LANG=C isosize -x "$DVD" | sed -r 's/.*: (.*), .*: (.*)$/\1 \2/')

    log_and_run 'Dump disc to ISO image' \
    dd -bs $sectsize count=$sectcount if="$DVD" of="$BUILD/.iso" || {
      log_and_run 'Dump disc to ISO image (slower because of media errors)' \
      ddrescue -n -b$sectsize "$DVD" "$BUILD/.iso" "$BUILD/.iso.mapfile"
      log_and_run 'Try to recover damaged sectors if needed' \
      ddrescue -d -r1 -b$sectsize "$DVD" "$BUILD/.iso" "$BUILD/.iso.mapfile"
    }
  fi
fi

# fetch metadata

if [ ! -f "$BUILD/.json" ]; then
  log_and_run -o 'Read disc metadata (tracks, chapters…)' \
  HandBrakeCLI --json -t 0 --min-duration 5 -i "$BUILD/.iso" 2>/dev/null \
  | sed $'1i \\\n{\n1,/JSON Title Set:/d' >"$BUILD/.json"
fi

ALL_META="$(
  jq -r ".TitleList[] | .Index as \$tnum | \
    \"TITLE \(\$tnum) (\(.Duration.Hours):\(.Duration.Minutes):\(.Duration.Seconds))\", \
    \"\tVideo (index, codec, geometry, frame-rate, bit-depth, chroma-subsampling, pixel aspect ratio, top:bottom:left:right borders, is interlaced)\", \
    \"V-\(\$tnum)\t\t0\t\(.VideoCodec)\t\(.Geometry.Width)x\(.Geometry.Height)\t\(.FrameRate.Num/.FrameRate.Den)\t\(.Color.BitDepth)\t\(.Color.ChromaSubsampling)\t\(1000*.Geometry.PAR.Num/.Geometry.PAR.Den | floor/1000)\t\(.Crop[0]):\(.Crop[1]):\(.Crop[2]):\(.Crop[3])\t\(.InterlaceDetected)\", \
    \"\tAudio (index, language, codec, channels, sample-rate, bit/s, is default, is commentary, is secondary, for visually impaired, description)\", \
    (.AudioList[] | \
      \"A-\(\$tnum)\t\t\(.TrackNumber)\t\(.LanguageCode)\t\(.CodecName)\t\(.ChannelCount)\t\(.SampleRate)\t\(.BitRate)\t\(.Attributes.Default)\t\(.Attributes.Commentary or .Attributes.AltCommentary)\t\(.Attributes.Secondary or .Attributes.AltCommentary)\t\(.Attributes.VisuallyImpaired)\t\(.Description)\"), \
    \"\tSubtitles (index, language, codec, is default, is commentary, is forced, for hearing impaired, description)\", \
    (.SubtitleList[] | \
      \"S-\(\$tnum)\t\t\(.TrackNumber)\t\(.LanguageCode)\t\(.SourceName):\(.Format)\t\(.Attributes.Default)\t\(.Attributes.Commentary)\t\(.Attributes.Forced)\t\(.Attributes.ClosedCaption)\t\(.Language)\"), \
    \"\tChapters (index, duration, name)\", \
    (.ChapterList | to_entries[] | \
      \"C-\(\$tnum)\t\t\(.key+1)\t\(.value.Duration.Hours):\(.value.Duration.Minutes):\(.value.Duration.Seconds)\t\(.value.Name)\")" \
    <"$BUILD/.json"
)"

# show information then exit if requested

if [ -n "$SHOW" ]; then
  echo "$ALL_META"
  [ -z "$KEEP_ALL" ] && rm -rf "$BUILD" || { [ -n "$RM_ISO" ] && rm -f "$BUILD/.iso"; }
  exit 0
fi

# parse information

function selectgroupby() {
  local subst=_
  if [ "$1" == _ ]; then subst=,; fi
  awk -vf=$1 -vg=$2 -vs="$3" -vr=$subst ${4:+-F"$4"} '
    $g!=group { if (group!="") { printf("\n") }; group=$g; gsub(s,r,$f); printf("%s %s",$g,$f); next }
    { gsub(s,r,$f); printf("%s%s",s,$f) }
    END { printf("\n") }'
}

declare -A VSECONDS VCROP
declare -A CHNAMES
declare -A ADEFAULT ALANGS ANUMS ACODS AMIXS ARATES ANAMES
declare -A SDEFAULT SLANGS SNUMS SNAMES

while read -r tnum h m s; do VSECONDS[$tnum]=$((3600*h+60*m+s)); done < <(
  sed -nr 's/^TITLE (.*) \((.*):(.*):(.*)\)/\1 \2 \3 \4/p' <<<"$ALL_META")
while read -r tnum chnum name; do CHNAMES[$tnum.$chnum]="$name"; done < <(
  sed -nr 's/^C-([0-9]+)\t\t([0-9]+)\t[0-9:]*\t(.*)/\1 \2 \3/p' <<<"$ALL_META")

while read -r tnum langs; do ALANGS[$tnum]="$langs"; done < <(
  sed -nr "s/^A-([0-9]+)\t\t[0-9]+\t(${ALL_LANGS//,/|})\t.*/\1 \2/p" <<<"$ALL_META" | selectgroupby 2 1 ,)
while read -r tnum nums; do ANUMS[$tnum]="$nums"; done < <(
  sed -nr "s/^A-([0-9]+)\t\t([0-9]+)\t(${ALL_LANGS//,/|})\t.*/\1 \2/p" <<<"$ALL_META" | selectgroupby 2 1 ,)
while read -r tnum cods; do ACODS[$tnum]="$cods"; done < <(
  sed -nr "
    s/^A-([0-9]+)\t\t[0-9]+\t(${ALL_LANGS//,/|})\t([^\t]*)\t.*/\1 \3/; T;
    s/ (aac|ac3|eac3|truehd|dts|dtshd|mp2|mp3|flac|opus)$/ copy:\1/p; t
    s/ .*/ flac16/p
  " <<<"$ALL_META" | selectgroupby 2 1 ,)
while read -r tnum mixs; do AMIXS[$tnum]="$mixs"; done < <(
  sed -nr "
    s/^A-([0-9]+)\t\t[0-9]+\t(${ALL_LANGS//,/|})\t[^\t]*\t([0-9]+)\t.*\t(.*)/\1 \3 \4/; T;
    s/ 1 .*/ mono/
    s/ 2 .*[Dd][Oo][Ll][Bb][Yy].*/ dpl2/
    s/ 2 .*/ stereo/
    s/ [345] .*/ dpl2/
    s/ 7 .*/ 6point1/
    s/ 8 .*/ 7point1/
    s/ [0-9]+ .*/ 5point1/
    p
  " <<<"$ALL_META" | selectgroupby 2 1 ,)
while read -r tnum rates; do ARATES[$tnum]="$rates"; done < <(
  sed -nr "s/^A-([0-9]+)\t\t[0-9]+\t(${ALL_LANGS//,/|})\t.*/\1 auto/p" <<<"$ALL_META" | selectgroupby 2 1 ,)
while read -r tnum names; do ANAMES[$tnum]="$names"; done < <(
  sed -nr "s/^A-([0-9]+)\t\t[0-9]+\t(${ALL_LANGS//,/|})\t.*\t(.*)/\1\t\3/p" <<<"$ALL_META" | selectgroupby 2 1 , $'\t')
while read -r tnum default; do ADEFAULT[$tnum]="${default%%,*}"; done < <(
  sed -nr "s/^A-([0-9]+)\t\t([0-9]+)\t$DEFAULT_LANG\t[^\t]*\t([0-9]+)(\t[^\t]*){2}((\t[^\t]*){4})\t.*/\1\t\2\t\3\5/p" <<<"$ALL_META" \
  | sort -t$'\t' -k1,1 -k5,5 -k7,7 -k6,6 -k3,3nr -k4,4r -k2,2n | selectgroupby 2 1 , $'\t')

while read -r tnum langs; do SLANGS[$tnum]="$langs"; done < <(
  sed -nr "s/^S-([0-9]+)\t\t[0-9]+\t(${ALL_LANGS//,/|})\t.*/\1 \2/p" <<<"$ALL_META" | selectgroupby 2 1 ,)
while read -r tnum nums; do SNUMS[$tnum]="$nums"; done < <(
  sed -nr "s/^S-([0-9]+)\t\t([0-9]+)\t(${ALL_LANGS//,/|})\t.*/\1 \2/p" <<<"$ALL_META" | selectgroupby 2 1 ,)
while read -r tnum names; do SNAMES[$tnum]="$names"; done < <(
  sed -nr "s/^S-([0-9]+)\t\t[0-9]+\t(${ALL_LANGS//,/|})\t.*\t(.*)/\1\t\3/p" <<<"$ALL_META" | selectgroupby 2 1 , $'\t')
while read -r tnum default; do SDEFAULT[$tnum]="${default%%,*}"; done < <(
  sed -nr "s/^S-([0-9]+)\t\t([0-9]+)\t$DEFAULT_LANG\t[^\t]*((\t[^\t]*){4})\t.*/\1\t\2\3/p" <<<"$ALL_META" \
  | sort -t$'\t' -k1,1 -k4,4 -k6,6 -k5,5r -k3,3r -k2,2n \
  | selectgroupby 2 1 , $'\t' \
  | grep -vf <(printf '^%d \n' "${!ADEFAULT[@]}")) # default subtitles only if no default audio was found

for tnum in "${!VSECONDS[@]}"; do
  if ! [ -f "$BUILD/.t.$tnum.tblr-crop" ]; then
    case ${VSECONDS[$tnum]} in
    ?|??) start=0; duration=$((${VSECONDS[$tnum]}-1)) ;;
    *)    start=$((${VSECONDS[$tnum]}/3)); duration=60 ;;
    esac
    IFS=: read -r cw ch cx cy < <(
      log_and_run -o "Detect black borders for title $tnum" \
      mplayer -dvd-device "$BUILD/.iso" -vo null -ao null -vf cropdetect -ss $start -endpos $duration -speed 10 dvd://$tnum 2>&1 \
      | grep -oe '-vf crop=.*[0-9]' | sort | uniq -c | sort -k1,1nr | head -n 1 | sed 's/.*=//')
    read -r rawx rawy < <(sed -nr "s/^V-$tnum\t(\t[^\t]*){2}\t([0-9]+)x([0-9]+)\t.*/\2 \3/p" <<<"$ALL_META")
    if [ -z "$cy" ] || [ -z "$rawy" ]; then
      touch "$BUILD/.t.$tnum.tblr-crop"
    else
      echo "--crop $cy:$((rawy-ch-cy)):$cx:$((rawx-cw-cx))" >"$BUILD/.t.$tnum.tblr-crop"
    fi
  fi
  VCROP[$tnum]="$(cat "$BUILD/.t.$tnum.tblr-crop")"
done

# rip chapters

if [ -z "$(ls "$BUILD/"t.*.ch.*.mkv 2>/dev/null)" ]; then
  for chapter in "${!CHNAMES[@]}"; do
    tnum=${chapter%.*}
    chnum=${chapter#*.}

    log_and_run "Rip track $tnum chapter $chnum (${CHNAMES[$chapter]})" \
    HandBrakeCLI -t $tnum --min-duration 5 -c $chnum --no-markers -e x265_10bit --encoder-profile main10 -q 18 --vfr \
      -a ${ANUMS[$tnum]:-none} -E ${ACODS[$tnum]:-none} --mixdown ${AMIXS[$tnum]:-none} -R ${ARATES[$tnum]:-none} -A "${ANAMES[$tnum]:-none}" \
      -X $MAXX -Y $MAXY ${VCROP[$tnum]} --auto-anamorphic --no-comb-detect --no-decomb --no-detelecine \
      --no-hqdn3d --no-nlmeans --no-chroma-smooth --no-unsharp --no-lapsharp --no-deblock \
      -s ${SNUMS[$tnum]:-none} -S "${SNAMES[$tnum]:-none}" --subtitle-burned=none \
      -i "$BUILD/.iso" -o "$BUILD/t.$tnum.ch.$chnum.mkv" </dev/null
  done
fi

unset ALANGS
unset ANUMS
unset ACODS
unset AMIXS
unset ARATES
unset ANAMES
unset SLANGS
unset SNUMS
unset SNAMES

# fetch chapter metadata

if [ -z "$(ls "$BUILD/"t.*.ch.*.json 2>/dev/null)" ]; then
  for mkv in "$BUILD/"t.*.ch.*.mkv; do
    log_and_run "Read metadata (length, codecs…) for $mkv" \
    mkvmerge -F json -i "$mkv" -r "${mkv%mkv}json"
  done
fi

declare -A STREAMS_IDS
declare -A CHAPTERS

jqscript='['
jqscript+='([.tracks[] | [select(.type=="video").properties] | sort_by(.number)[] | "\(.number),\(.codec_id),\(.pixel_dimensions)"] | join("+"))'
jqscript+=',([.tracks[] | [select(.type=="audio").properties] | sort_by(.number)[] | "\(.number),\(.codec_id),\(.audio_channels)x\(.audio_sampling_frequency),\(.language)"] | join("+"))'
jqscript+=',([.tracks[] | [select(.type=="subtitles").properties] | sort_by(.number)[] | "\(.number),\(.codec_id),\(.language)"] | join("+"))'
jqscript+='] | join("|")'
TITLE_LIST=''
while read -r tnum chnum; do
  TITLE_LIST="${TITLE_LIST% $tnum} $tnum"
  CHAPTERS[$tnum]="${CHAPTERS[$tnum]} $chnum"
done < <(ls -1 "$BUILD/"t.*.ch.*.json | sed -r 's#.*/t\.(.*)\.ch\.(.*)\.json#\1 \2#' | sort -k1,1n -k2,2n)
for tnum in $TITLE_LIST; do
  read chnum x <<<"${CHAPTERS[$tnum]}"
  STREAMS_IDS[$tnum]="$(jq -r "$jqscript" <"$BUILD/t.$tnum.ch.$chnum.json")"
done

# generate per-chapter timeline snapshots

if [ -z "$(ls "$BUILD/"t.*.ch.*.png 2>/dev/null)" ]; then
  [ -d "$BUILD/.tsnap" ] || mkdir "$BUILD/.tsnap"
  while read -r tnum chnum x; do
    seconds=$(jq -r '.container.properties.duration / .container.properties.timestamp_scale / 1000 | floor' <"$BUILD/t.$tnum.ch.$chnum.json")
    montagecmd=(montage -background black)
    snapsec=0
    snapcount=0
    while [ $snapsec -le $seconds ]; do
      target="$BUILD/.tsnap/t.$tnum.ch.$chnum.s.$snapsec.png"
      log_and_run "Make snapshot @${snapsec}s of track $tnum chapter $chnum" \
      ffmpeg -i "$BUILD/t.$tnum.ch.$chnum.mkv" -ss ${snapsec}s -vf scale=w=256:h=144:force_original_aspect_ratio=decrease -vframes 1 "$target%1d.png" </dev/null
      mv "$target"*.png "$target"
      snapsec=$((snapsec+5))
      snapcount=$((snapcount+1))
      montagecmd+=("$target")
    done
    cellsize=$(identify "$BUILD/.tsnap/t.$tnum.ch.$chnum.s.5.png" </dev/null | sed -nr 's/.* ([0-9]+x[0-9]+(\+[0-9]+){2}).*/\1/p')
    gridsize=1
    while [ $((gridsize*gridsize)) -lt $snapcount ]; do gridsize=$((gridsize+1)); done
    log_and_run "Make timeline snapshot of track $tnum chapter $chnum" \
    "${montagecmd[@]}" -tile ${gridsize}x${gridsize} -geometry $cellsize -trim "$BUILD/.tsnap/t.$tnum.ch.$chnum.png" </dev/null
    log_and_run "Annotate timeline snapshot of track $tnum chapter $chnum with audio and subtitles metadata" \
    convert -fill red -gravity SouthEast -annotate +0+0 "${STREAMS_IDS[$tnum]}" "$BUILD/.tsnap/t.$tnum.ch.$chnum.png" "$BUILD/t.$tnum.ch.$chnum.png" </dev/null
  done < <(
    # restrict to chapters having common durations
    jq -r '.TitleList[] | .Index as $tnum | .ChapterList | to_entries | .[] | .key as $chnum | .value.Duration | "\($tnum) \(1+$chnum) \(.Hours).\(.Minutes).\(.Seconds).\(.Ticks)"' <"$BUILD/.json" \
    | sort -k3,3V | uniq -f2 -D # keep only duplicates in the duration column
  )
fi

# detect duplicates

function islongfilm() {
  [ ${VSECONDS[$1]} -ge 4200 ] # [70min–…]
  echo $?
}

function isepisode() {
  [ ${VSECONDS[$1]} -ge 900 ] # [15min–…]
  echo $?
}

if [ ! -f "$BUILD/.duplicates" ]; then
  log_and_run -o "Detect duplicates among all chapters of all titles" \
  czkawka_cli image -f "$BUILD/.dedup" -s High -m 1 -R -d "$BUILD/" \
  | sed -rn 's#^/.*/t\.([0-9]+)\.ch\.([0-9]+)\.png.*#\1 \2#p' \
  | while read -r tnum1 chnum1; read -r tnum2 chnum2; do
    if [ "${STREAMS_IDS[$tnum1]}" == "${STREAMS_IDS[$tnum2]}" ]; then
      # long film first, else episode first, else lower title number first, else lower chapter number first
      printf '%d %d %d %d %d %d\n%d %d %d %d %d %d\n' \
        $(islongfilm $tnum1) $(isepisode $tnum1) $tnum1 $tnum2 $chnum1 $chnum2 \
        $(islongfilm $tnum2) $(isepisode $tnum2) $tnum2 $tnum1 $chnum2 $chnum1 \
      | sort -k1,1n -k2,2n -k3,3n -k4,4n -k5,5n | head -n 1 | awk '{print $3, $5, $4, $6}'
    fi
  done >"$BUILD/.duplicates"
fi

# detect source tracks that will fit together within the same end-result track

if [ ! -f "$BUILD/.titlelist" ]; then
  for tnum in $TITLE_LIST; do
    strid="${STREAMS_IDS[$tnum]}"
    lowestwithid=$(for i in "${!STREAMS_IDS[@]}"; do printf '%d\t%s\t\n' $i "${STREAMS_IDS[$i]}"; done \
      | grep -F $'\t'"$strid"$'\t' | sort -t$'\n' -k1,1n | head -n 1 | cut -f1)
    printf '%d %d %d %d\n' $(islongfilm $tnum) $(isepisode $tnum) $lowestwithid $tnum
  done \
  | sort -k1,1n -k2,2n -k3,3n -k4,4n | awk '
    $1=="0"{print $4; next}
    ($2 $3)!=streams{if(streams!=""){printf("\n")};streams=($2 $3);printf("%d",$4);next}
    {printf(" %d",$4)}
    END{printf("\n")}
  ' >"$BUILD/.titlelist"
fi

# compute chapters’ timestamps, UIDs, and substitutions + prepare merging

declare -A TIMESTAMPS
declare -a STREAMS_LIST
declare -A STREAMS_TO_REF
declare -A STREAMS_TO_FILES
declare -A CHAPTERS_IDS
#declare -A UIDS
declare -A SUBST
declare -a MKVMRGARGS

while read -r t1 ch1 t2 ch2; do
  SUBST[$t2.$ch2]="$t1.$ch1"
done <"$BUILD/.duplicates"

while read -r tlist; do
  strid="${STREAMS_IDS[${tlist%% *}]}"
  if printf '%s\n' "${STREAMS_LIST[@]}" | grep -qxF "$strid"; then
    ref=${STREAMS_TO_REF[$strid]}
  else
    ref=0
    STREAMS_LIST+=("$strid")
  fi
  for tnum in $tlist; do
    for chnum in ${CHAPTERS[$tnum]}; do
      tch=$tnum.$chnum
      while [ -n "${SUBST[$tch]}" ]; do tch=${SUBST[$tch]}; done
      if [ $tch == $tnum.$chnum ]; then
        # compute milliseconds
        start=$ref
        stop=$(bc -lq <<<"$ref+$(jq -r '.container.properties | .duration/.timestamp_scale' <"$BUILD/t.$tnum.ch.$chnum.json")")
        ref=$stop
        TIMESTAMPS[$tch]="$start $stop"
        #UIDS[$tch]="$(jq -r .container.properties.segment_uid <"$BUILD/t.$tnum.ch.$chnum.json")"
        STREAMS_TO_FILES[$strid]="${STREAMS_TO_FILES[$strid]}|$BUILD/t.$tnum.ch.$chnum.mkv"
      else
        TIMESTAMPS[$tnum.$chnum]="${TIMESTAMPS[$tch]}"
        #UIDS[$tnum.$chnum]="${UIDS[$tch]}"
      fi
      CHAPTERS_IDS[$tnum.$chnum]="$(jq -r .container.properties.segment_uid <"$BUILD/t.$tnum.ch.$chnum.json" \
        | tr '[a-f]' '[A-F]' | bc -q <<<"ibase=16;$(cat)" | head -c 10)"
    done
  done
  STREAMS_TO_REF[$strid]=$ref
done <"$BUILD/.titlelist"

declare -A STREAMS_TO_OFFSET STREAMS_TO_TRACKID

ref=0
id=1
for strid in "${STREAMS_LIST[@]}"; do
  STREAMS_TO_OFFSET[$strid]=$ref
  STREAMS_TO_TRACKID[$strid]=$id; id=$((id+1))
  file="${STREAMS_TO_FILES[$strid]##*|}"
  tnum=$(sed -r 's#.*/t.([0-9]+).ch.[0-9]+.mkv#\1#' <<<"$file")
  streamcount=$(jq -r '.tracks | length' <"${file%mkv}json")
  for ((i=0; i<streamcount; i++)); do
    MKVMRGARGS+=('--sync' $i:$ref --default-track-flag)
    if [ $i == "${ADEFAULT[$tnum]}" ] || [ $i == "${SDEFAULT[$tnum]}" ]; then
      MKVMRGARGS+=($i:1)
    else
      MKVMRGARGS+=($i:0)
    fi
  done
  MKVMRGARGS+=('--no-chapters' '--no-global-tags' '[')
  IFS='|' read -r -a files <<<"${STREAMS_TO_FILES[$strid]#|}"
  MKVMRGARGS+=("${files[@]}")
  MKVMRGARGS+=(']')
#  for file in "${files[@]}"; do
#    ref=$(bc -lq <<<"$ref+$(jq -r '.container.properties | .duration/.timestamp_scale' <"${file%mkv}json")")
#  done
done

# generate MKV editions XML file

function formatTimestamp() {
  if [ $1 == 0 ]; then
    echo '00:00:00.000'
  else
    date -u -d@${1:0:-3} +%T.${1: -3:3}
  fi
}

if [ ! -f "$BUILD/mkv-editions.xml" ]; then
{
  cat <<ENDOFXML
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE Chapters SYSTEM "matroskachapters.dtd">
<Chapters>
ENDOFXML

  edflagdefault=1
  titlenum=1
  while read -r titlelist; do
    strid="${STREAMS_IDS[${titlelist%% *}]}"
    offset=${STREAMS_TO_OFFSET[$strid]}

    cat <<ENDOFXML
  <EditionEntry>
    <EditionFlagHidden>0</EditionFlagHidden>
    <EditionFlagDefault>$edflagdefault</EditionFlagDefault>
    <EditionFlagOrdered>1</EditionFlagOrdered>
    <EditionDisplay>
      <EditionString>$(printf '%03d' $titlenum): $titlelist ($(tr '|+' ' ' <<<"$strid"))</EditionString>
    </EditionDisplay>
ENDOFXML

    chflaghidden=0
    for tnum in $titlelist; do
      for chnum in ${CHAPTERS[$tnum]}; do

        cat <<ENDOFXML
    <ChapterAtom>
      <ChapterUID>${CHAPTERS_IDS[$tnum.$chnum]}</ChapterUID>
      <ChapterTimeStart>$(formatTimestamp $(printf '%s+%s\n' $offset ${TIMESTAMPS[$tnum.$chnum]% *} | bc -q))</ChapterTimeStart>
      <ChapterTimeEnd>$(formatTimestamp $(printf '%s+%s\n' $offset ${TIMESTAMPS[$tnum.$chnum]#* } | bc -q))</ChapterTimeEnd>
      <ChapterFlagHidden>$chflaghidden</ChapterFlagHidden>
      <ChapterFlagEnabled>1</ChapterFlagEnabled>
      <!-- ChapterSegmentUID format="hex">$ - {UIDS[$tnum.$chnum]} - </ChapterSegmentUID -->
      <ChapterDisplay>
        <ChapterString>$tnum.$chnum — ${CHNAMES[$tnum.$chnum]}</ChapterString>
      </ChapterDisplay>
      <ChapterTrack>
        <ChapterTrackNumber>${STREAMS_TO_TRACKID[$strid]}</ChapterTrackNumber>
      </ChapterTrack>
    </ChapterAtom>
ENDOFXML

        if [ "${titlelist/ }" != "$titlelist" ]; then
          chflaghidden=1
        fi
      done
      chflaghidden=0
    done

    cat <<ENDOFXML
  </EditionEntry>
ENDOFXML

    edflagdefault=0
    titlenum=$((titlenum+1))
  done <"$BUILD/.titlelist"

  cat <<ENDOFXML
</Chapters>
ENDOFXML

} >"$BUILD/mkv-editions.xml"
fi

# merge all chapters of all titles

if [ ! -f "$TARGET" ]; then
  log_and_run 'Build final MKV file' \
  mkvmerge --disable-track-statistics-tags --append-mode file --chapters "$BUILD/mkv-editions.xml" -o "$TARGET" "${MKVMRGARGS[@]}"
fi

if [ $? -eq 0 ] && [ -z "$KEEP_ALL" ]; then
  rm -rf "$BUILD"
fi
